@@ -25,15 +25,18 @@
 #  Frédéric MARTIN frederic.martin.fma@gmail.com
 #
 
+from __future__ import unicode_literals
+
 import os
 import sys
-import io
 import logging
 import base64
 import copy
 from ordereddict import OrderedDict
 import ConfigParser
 import argparse
+import types
+import locale
 
 # -----------------------------------------------------------------------------
 # global logger variable
@@ -92,12 +95,12 @@ class Base64ElementHook(DefaultHook):
                 log = logging.getLogger('argtoolbox')
                 if self.warning:
                     log.warn("current field '%(name)s' is not \
-                        stored in the configuration file with \
-                        base64 encoding",
+stored in the configuration file with \
+base64 encoding",
                              {"name": getattr(elt, "_name")})
                 else:
                     log.error("current field '%(name)s' is not stored in the \
-                    configuration file with base64 encoding", {"name":
+configuration file with base64 encoding", {"name":
                               getattr(elt, "_name")})
                     raise ex
 
@@ -118,12 +121,12 @@ class SectionHook(object):
             raise TypeError("First argument should be a subclass of _Section.")
         self.section = section
 
-        if not isinstance(attribute, str):
+        if not isinstance(attribute, types.UnicodeType):
             raise TypeError("Second argument should be a string, "
                             + "attribute name.")
         self.attribute = attribute
 
-        if not isinstance(opt_name, str):
+        if not isinstance(opt_name, types.UnicodeType):
             raise TypeError("Third argument should be a string, option name.")
         self.opt_name = opt_name
 
@@ -152,7 +155,7 @@ class Config(object):
         self.sections = OrderedDict()
         self._default_section = self.add_section(SimpleSection("DEFAULT"))
         self.parser = None
-        self.fileParser = ConfigParser.SafeConfigParser()
+        self.file_parser = ConfigParser.SafeConfigParser()
 
     def add_section(self, section):
         """Add a new Section object to the config. Should be a subclass of
@@ -179,11 +182,12 @@ class Config(object):
         log = logging.getLogger('argtoolbox')
         discoveredFileList = []
         if self.config_file:
-            if isinstance(self.config_file, str):
-                discoveredFileList = self.fileParser.read(self.config_file)
+            if isinstance(self.config_file, types.UnicodeType):
+                discoveredFileList = self.file_parser.read(self.config_file)
             else:
-                discoveredFileList = self.fileParser.readfp(self.config_file,
-                                                            "file descriptor")
+                discoveredFileList = self.file_parser.readfp(
+                    self.config_file,
+                    "file descriptor")
         else:
             defaultFileList = []
             defaultFileList.append(self.prog_name + ".cfg")
@@ -191,7 +195,7 @@ class Config(object):
                 os.path.expanduser('~/.' + self.prog_name + '.cfg'))
             defaultFileList.append('/etc/' + self.prog_name + '.cfg')
             log.debug("defaultFileList: " + str(defaultFileList))
-            discoveredFileList = self.fileParser.read(defaultFileList)
+            discoveredFileList = self.file_parser.read(defaultFileList)
 
         log.debug("discoveredFileList: " + str(discoveredFileList))
 
@@ -206,13 +210,13 @@ class Config(object):
             for s in self.sections.values():
                 log.debug("loading section : " + s.get_section_name())
                 try:
-                    s.load(self.fileParser)
+                    s.load(self.file_parser)
                 except ValueError:
                     sys.exit(1)
         else:
             for s in self.sections.values():
                 log.debug("loading section : " + s.get_section_name())
-                s.load(self.fileParser)
+                s.load(self.file_parser)
 
         log.debug("configuration loaded.")
 
@@ -244,7 +248,15 @@ class Config(object):
         # Parsing the command line looking for the previous options like
         # configuration file name or server section. Extra arguments
         # will be store into argv.
-        args = self.parser.parse_known_args()[0]
+        args = None
+        if os.environ.get('_ARGCOMPLETE'):
+            # During argcomplete completion, parse_known_args will return an
+            # empty Namespace. In this case, we feed the previous function with
+            # data comming from the input completion data
+            compline = os.environ.get('COMP_LINE')
+            args = self.parser.parse_known_args(compline.split()[1:])[0]
+        else:
+            args = self.parser.parse_known_args()[0]
 
         if hooks is not None:
             if isinstance(hooks, list):
@@ -267,10 +279,10 @@ class Config(object):
             log = logging.getLogger('argtoolbox')
             log.debug("reloading configuration ...")
             if args.config_file:
-                self.fileParser.read(args.config_file)
+                self.file_parser.read(args.config_file)
             for s in self.sections.values():
                 log.debug("loading section : " + s.get_section_name())
-                s.load(self.fileParser)
+                s.load(self.file_parser)
             log.debug("configuration reloaded.")
 
     def __getattr__(self, name):
@@ -281,7 +293,7 @@ class Config(object):
             return s
         else:
             raise AttributeError("'%(class)s' object has no attribute \
-            '%(name)s'" % {"name": name, "class": self.__class__.__name__})
+'%(name)s'" % {"name": name, "class": self.__class__.__name__})
 
     def __str__(self):
         res = []
@@ -303,7 +315,10 @@ class Config(object):
                     f.write("#####################################\n")
                     f.write("# Description :\n")
                     f.write("# -------------\n# ")
-                    f.write(self._desc)
+                    for i in self._desc.split('\n'):
+                        f.write("# ")
+                        f.write(i)
+                        f.write("\n")
                     f.write("\n\n")
 
                 for s in self.sections.values():
@@ -346,7 +361,7 @@ class _AbstractSection(object):
 
     # pylint: disable-msg=W0613
     # pylint: disable-msg=R0201
-    def load(self, fileParser):
+    def load(self, file_parser):
         """ This method must be implemented by the subclass. This method should
         read and load all section elements.
         """
@@ -367,12 +382,16 @@ class _AbstractSection(object):
         """
         if comments:
             f.write("#####################################\n")
-            f.write("# Section : " + "#".join(self.get_representation()) + "\n")
+            f.write("# Section : ")
+            f.write("#".join(self.get_representation()) + "\n")
             f.write("#####################################\n")
         f.write("[" + self._name + "]\n")
         if self._desc and comments:
             f.write("# Description : ")
-            f.write(self._desc)
+            for i in self._desc.split('\n'):
+                f.write("# ")
+                f.write(i)
+                f.write("\n")
             f.write("\n")
 
 
@@ -402,11 +421,11 @@ class _Section(_AbstractSection):
         Section"""
         return len(self.elements)
 
-    def load(self, fileParser):
+    def load(self, file_parser):
         section = self.get_section_name()
         try:
             for e in self.elements.values():
-                e.load(fileParser, section)
+                e.load(file_parser, section)
         except ConfigParser.NoSectionError as e:
             # pylint: disable-msg=W0621
             log = logging.getLogger('argtoolbox')
@@ -422,7 +441,7 @@ class _Section(_AbstractSection):
             return e
         else:
             raise AttributeError("'%(class)s' object has no attribute \
-            '%(name)s'" % {"name": name, "class": self.__class__.__name__})
+'%(name)s'" % {"name": name, "class": self.__class__.__name__})
 
     def write_config_file(self, f, comments):
         """This method write a sample file, with attributes, descriptions,
@@ -498,13 +517,15 @@ class ListSection(_AbstractSection):
         self.elements = OrderedDict()
         self._name = name
 
-    def load(self, fileParser):
+    def load(self, file_parser):
 
         section = self.get_section_name()
         try:
-            for key in [item for item in fileParser.options(section)
-                        if item not in fileParser.defaults().keys()]:
-                self.elements[key] = fileParser.get(section, key)
+
+            # TODO : ? : data = data.decode(locale.getpreferredencoding())
+            for key in [item for item in file_parser.options(section)
+                        if item not in file_parser.defaults().keys()]:
+                self.elements[key] = file_parser.get(section, key)
         except ConfigParser.NoSectionError as e:
             # pylint: disable-msg=W0621
             log = logging.getLogger('argtoolbox')
@@ -630,8 +651,14 @@ class Element(object):
             res.append(prefix + " - " + str(self._name)
                        + " : xxxxxxxx" + suffix)
         else:
-            res.append(prefix + " - " + str(self._name)
-                       + " : " + str(self.value) + suffix)
+            a = prefix + " - "
+            a += str(self._name) + " : "
+            if isinstance(self.value, types.UnicodeType):
+                a += self.value
+            else:
+                a += str(self.value)
+            a += suffix
+            res.append(a)
         return res
 
     def __str__(self):
@@ -650,15 +677,15 @@ class Element(object):
         for h in self.hooks:
             h(self)
 
-    def load(self, fileParser, section_name):
+    def load(self, file_parser, section_name):
         """The current element is loaded from the configuration file,
         all constraints and requirements are checked.
         Then element hooks are applied.
         """
-        self._load(fileParser, section_name)
+        self._load(file_parser, section_name)
         self.post_load()
 
-    def _load(self, fileParser, section_name):
+    def _load(self, file_parser, section_name):
         """The current element is loaded from the configuration file,
         all constraints and requirements are checked.
         """
@@ -670,22 +697,24 @@ class Element(object):
             data = None
             try:
                 if self.e_type == int:
-                    data = fileParser.getint(section_name, self._name)
+                    data = file_parser.getint(section_name, self._name)
                 elif self.e_type == float:
-                    data = fileParser.getfloat(section_name, self._name)
+                    data = file_parser.getfloat(section_name, self._name)
                 elif self.e_type == bool:
-                    data = fileParser.getboolean(section_name, self._name)
+                    data = file_parser.getboolean(section_name, self._name)
                 elif self.e_type == list:
-                    data = fileParser.get(section_name, self._name)
-                    data = data.strip().split()
+                    data = file_parser.get(section_name, self._name)
+                    data = data.strip()
+                    data = data.decode(locale.getpreferredencoding())
+                    data = data.split()
                     if not data:
                         msg = "The optional field '%(name)s' was present, \
-                        type is list, but the current value is an empty \
-                        list." % {"name": self._name}
+type is list, but the current value is an empty \
+list." % {"name": self._name}
                         log.error(msg)
                         raise ValueError(msg)
                 elif self.e_type == str:
-                    data = fileParser.get(section_name, self._name)
+                    data = file_parser.get(section_name, self._name)
                     # happens only when the current field is present,
                     # type is string, but value is ''
                     if not data:
@@ -694,15 +723,16 @@ class Element(object):
                         string." % {"name": self._name}
                         log.error(msg)
                         raise ValueError(msg)
+                    data = data.decode(locale.getpreferredencoding())
                 else:
-                    msg = "Data type not supported : %(type)s\
-                    " % {"type": self.e_type}
+                    msg = "Data type not supported : %(type)s " % {
+                          "type": self.e_type}
                     log.error(msg)
                     raise TypeError(msg)
 
             except ValueError as ex:
                 msg = "The current field '%(name)s' was present, but the \
-                required type is : %(e_type)s." % {
+required type is : %(e_type)s." % {
                     "name": self._name,
                     "e_type": self.e_type
                     }
@@ -715,13 +745,13 @@ class Element(object):
             if self.hidden:
                 log_data['data'] = "xxxxxxxx"
             log.debug("field found : '%(name)s', value : '%(data)s', \
-                        type : '%(e_type)s'", log_data)
+type : '%(e_type)s'", log_data)
             self.value = data
 
         except ConfigParser.NoOptionError:
             if self.conf_required:
                 msg = "The required field '%(name)s' was missing from the \
-                config file." % {"name": self._name}
+config file." % {"name": self._name}
                 log.error(msg)
                 raise ValueError(msg)
 
@@ -732,7 +762,7 @@ class Element(object):
                 if self.hidden:
                     log_data['data'] = "xxxxxxxx"
                 log.debug("Field not found : '%(name)s', default value : \
-                    '%(data)s', type : '%(e_type)s'", log_data)
+'%(data)s', type : '%(e_type)s'", log_data)
             else:
                 log.debug("Field not found : '" + self._name + "'")
 
@@ -775,7 +805,10 @@ class Element(object):
             f.write("\n")
             if self._desc and self._desc != argparse.SUPPRESS:
                 f.write("# Description : ")
-                f.write(self._desc)
+                for i in self._desc.split('\n'):
+                    f.write("# ")
+                    f.write(i)
+                    f.write("\n")
                 f.write("\n")
 
         if not self.conf_required:
@@ -832,12 +865,12 @@ class ElementWithSubSections(Element):
         self.sections[section.name] = section
         return section
 
-    def load(self, fileParser, section_name):
-        self._load(fileParser, section_name)
+    def load(self, file_parser, section_name):
+        self._load(file_parser, section_name)
         if len(self.sections) > 0:
             for sec in self.sections.values():
                 sec.name = self.value
-                sec.load(fileParser)
+                sec.load(file_parser)
         self.post_load()
 
 
@@ -897,15 +930,15 @@ class ElementWithRelativeSubSection(ElementWithSubSections):
                             not :" + str(_Section.__class__))
         self.rss = rss
 
-    def load(self, fileParser, section_name):
-        self._load(fileParser, section_name)
+    def load(self, file_parser, section_name):
+        self._load(file_parser, section_name)
         if isinstance(self.value, list):
             for sec_name in self.value:
                 try:
                     sec = copy.deepcopy(self.rss, None)
                     setattr(sec, '_name', sec_name)
                     self.sections[sec_name] = sec
-                    sec.load(fileParser)
+                    sec.load(file_parser)
                 except ValueError as e:
                     # pylint: disable-msg=W0621
                     log = logging.getLogger('argtoolbox')
@@ -954,12 +987,16 @@ class DefaultCommand(object):
                 setattr(dict_tmp, field, "xxxxxxxx")
         self.log.debug("Namespace : begin :")
         for i in dict_tmp.__dict__:
-            self.log.debug(i + " : " + str(getattr(dict_tmp, i)))
+            attribute = getattr(dict_tmp, i)
+            if isinstance(attribute, types.UnicodeType):
+                self.log.debug(i + " : " + attribute)
+            else:
+                self.log.debug(i + " : " + str(attribute))
         self.log.debug("Namespace : end.")
 
     # pylint: disable-msg=W0613
     # pylint: disable-msg=R0201
-    def complete(self, args,  prefix):
+    def complete(self, args, prefix):
         """Auto complete method, args is comming from argparse and prefix is
         the input data from command line.
         You must return a list."""
@@ -978,7 +1015,7 @@ class TestCommand(DefaultCommand):
 
         print "The loaded configuration : "
         print "---------------------------"
-        print self.config
+        print unicode(self.config)
         print ""
 
         print "The command line arguments (argv) : "
@@ -1000,22 +1037,21 @@ class DefaultCompleter(object):
     def __call__(self, prefix, **kwargs):
         from argcomplete import debug
         try:
-            debug("\n-----------------------------")
-            debug(str(kwargs))
+            debug("\n------------ DefaultCompleter -----------------")
+            debug("Kwargs content :")
             for i, j in kwargs.items():
-                debug(i)
-                debug("\t" + str(j))
+                debug("key : " + str(i))
+                debug("\t - " + str(j))
+            debug("\n------------ DefaultCompleter -----------------\n")
 
             args = kwargs.get('parsed_args')
             # pylint: disable-msg=W0612
-            l_parser = kwargs.get('parser')
+            parser = kwargs.get('parser')
             #a = parser.parse_known_args()
-            debug("\n-----------------------------")
-            debug(str(args))
 
             # getting form args the current Command and looking for a method
-            # called by default 'complete'.
-            # The method name is specified  by func_name
+            # called by default 'complete'. See __init__ method. The method
+            # name is store in the class member called self.func_name
             fn = getattr(args.__func__, self.func_name, None)
             if fn:
                 return fn(args, prefix)
@@ -1029,7 +1065,7 @@ class DefaultCompleter(object):
 class DefaultProgram(object):
     """ TODO """
 
-    def __init__(self, parser, config = None):
+    def __init__(self, parser, config=None):
         self.parser = parser
         self.config = config
 
@@ -1049,8 +1085,8 @@ class DefaultProgram(object):
             llog.setLevel(logging.DEBUG)
             streamHandler.setFormatter(DEBUG_LOGGING_FORMAT)
             if self.config:
-                print "debug>>>----------- config ------------------"
-                print self.config
+                print "debug>>>----------- config -------------------"
+                print unicode(self.config)
                 print "debug----------- processing --------------<<<<"
 
             # run command
@@ -1077,15 +1113,16 @@ class BasicProgram(object):
     """ TODO """
 
     def __init__(self, name, config_file=None, desc=None,
-                 mandatory=False, use_config_file=True):
+                 mandatory=False, use_config_file=True, version="0.1-alpha"):
 
         # create configuration
-        self.config=Config(name, config_file=config_file, desc=desc,
+        self.config = Config(name, config_file=config_file, desc=desc,
                              mandatory=mandatory,
                              use_config_file=use_config_file)
         self.parser = None
-
+        self.version = version
         self.log = self.init_logger()
+        self.formatter_class = None
 
     def init_logger(self):
         # logger
@@ -1112,26 +1149,37 @@ class BasicProgram(object):
                     default=0,
                     desc="""debug level : default : 0."""))
 
+    def load(self):
+        # loading default configuration from the file
+        self.config.load()
+
     def init_parser(self):
         # arguments parser
         self.parser = self.config.get_parser()
+        if self.formatter_class:
+            self.parser.formatter_class = self.formatter_class
         self.parser.add_argument('-v', '--verbose',
                                  action="store_true",
                                  default=False)
         self.parser.add_argument('--version',
                                  action="version",
-                                 version="%(prog)s 0.1")
+                                 version="%(prog)s " + self.version)
 
-    def add_pre_arguments(self):
+    def add_pre_commands(self):
         """ You can override this method in order to add your command line
         arguments to the argparse parser. The configuration file is already
         loaded at this time."""
         pass
 
+    def reload(self):
+        # reloading configuration with previous optional arguments
+        # (example : config file name from argv, ...)
+        self.config.reload()
+
     def add_commands(self):
         """ You can override this method in order to add your command line
-        arguments to the argparse parser. The configuration file is already
-        loaded at this time."""
+        arguments to the argparse parser. The configuration file  was
+        reloaded at this time."""
         self.parser.add_argument(
             '-d',
             action="count",
@@ -1143,18 +1191,18 @@ class BasicProgram(object):
         self.add_config_options()
 
         # loading default configuration from the file
-        self.config.load()
+        self.load()
 
         # initialisation of the cli parser,
         # some default arguments are also added.
         self.init_parser()
 
         # adding some user arguments
-        self.add_pre_arguments()
+        self.add_pre_commands()
 
         # reloading configuration with previous optional arguments
         # (example : config file name from argv, ...)
-        self.config.reload()
+        self.reload()
 
         # adding all commands
         self.add_commands()
@@ -1213,94 +1261,3 @@ def _query_yes_no(question, default="yes"):
         else:
             sys.stdout.write("Please respond with 'yes' or 'no' "
                              "(or 'y' or 'n').\n")
-
-
-# -----------------------------------------------------------------------------
-# MAIN
-# -----------------------------------------------------------------------------
-if __name__ == "__main__":
-
-    sample_config = """
-[ldap]
-
-host=127.0.0.1
-port=389
-suffix=dc=nodomain
-account=cn=admin,dc=nodomain
-password=toto
-
-\n"""
-
-    # logger
-    glog = logging.getLogger()
-    glog.setLevel(logging.INFO)
-    # logger handlers
-    glog.addHandler(streamHandler)
-    # debug mode
-    # if you need debug during class construction, file config loading, ...,
-    # you need to modify the logger level here.
-    #glog.setLevel(logging.DEBUG)
-    #streamHandler.setFormatter(DEBUG_LOGGING_FORMAT)
-
-    # create configuration
-    myconfig = Config("sample-program",
-                      config_file=io.BytesIO(sample_config),
-                      desc="""Just a description for a sample program.
-This program supports argcomplete.
-To enable it, run in bash terminal:
-    eval "$(register-python-argcomplete argtoolbox.py)"
-""")
-
-    # section ldap
-    section_ldap = myconfig.add_section(SimpleSection("ldap"))
-    section_ldap.add_element(Element('debug',
-                                     e_type=int,
-                                     default=0,
-                                     desc="""debug level : default : 0."""))
-    section_ldap.add_element(Element('host',
-                                     required=True,
-                                     default="192.168.1.1"))
-    section_ldap.add_element(Element('account', required=True))
-    section_ldap.add_element(Element('port', e_type=int))
-    section_ldap.add_element(Element('password',
-                                     required=True,
-                                     hidden=True,
-                                     desc="account password to ldap",
-                                     hooks=[Base64ElementHook(), ]))
-
-    # loading default configuration
-    myconfig.load()
-
-    # -------------------------------------------------------------------------
-    # arguments parser
-    # -------------------------------------------------------------------------
-    myparser = myconfig.get_parser(
-        formatter_class=argparse.RawTextHelpFormatter)
-    # pylint: disable-msg=W0142
-    myparser.add_argument('-d', action="count",
-                          **myconfig.ldap.debug.get_arg_parse_arguments())
-    myparser.add_argument('-v', '--verbose',
-                          action="store_true",
-                          default=False)
-    myparser.add_argument('--version',
-                          action="version",
-                          version="%(prog)s 0.1")
-
-    # reloading configuration with previous optional arguments
-    # (example : config file name from argv, ...)
-    myconfig.reload()
-
-    # Adding all others parsers.
-    subparsers = myparser.add_subparsers()
-    parser_tmp = subparsers.add_parser(
-        'test',
-        help="This simple command print cli argv and configuration read \
-        form config file.")
-    parser_tmp.set_defaults(__func__=TestCommand(myconfig))
-
-    # run
-    prog = DefaultProgram(myparser)
-    if prog():
-        sys.exit(0)
-    else:
-        sys.exit(1)
